from gen3.submission import Gen3Submission
import json

from pathlib import Path


class MetadataExporter(object):
    """
    Class for exporting Gen3 metadata
    """
    def __init__(self, auth):
        """
        Constructor

        :param auth: Gen3 authentication object created by the Auth class
        :type auth: object
        """
        self._auth = auth
        self._exporter = Gen3Submission(self._auth)

    def export_node(self, program, project, node_type, fileformat, filename=None):
        """
        Exporting all records in a single Gen3 node

        :param program: Program name
        :type program: str
        :param project: Project
        :type project: str
        :param node_type: Node name
        :type node_type: str
        :param fileformat: Exported file format (json or tsv)
        :type fileformat: str
        :param filename: Exported filename
        :type filename: str
        :return: List of records (metadata) in dictionary format
        :rtype: list
        """
        output = self._exporter.export_node(program, project, node_type, fileformat, filename)
        data = json.loads(output).get("data")
        return data

    def export_record(self, program, project, uuid, fileformat, filename=None):
        """
        Exporting the metadata in a single record

        :param program: Program name
        :type program: str
        :param project: Project
        :type project: str
        :param uuid: Record UUID
        :type uuid: str
        :param fileformat: Exported file format (json or tsv)
        :type fileformat: str
        :param filename: Exported filename
        :type filename: str
        :return: Metadata in a single record
        :rtype: dict
        """
        output = self._exporter.export_record(program, project, uuid, fileformat, filename)
        data = json.loads(output)[0]
        return data

    @staticmethod
    def save_metadata(data, path, format="json"):
        """
        Saving the metadata (dict) in json format

        :param data: metadata
        :type data: dict
        :param format: file format (currently only json)
        :type format: str
        :param path: Path the save file
        :type path: str
        :return:
        :rtype:

        """
        path = Path(path)
        # Export data as either 'json' or 'tsv'
        if format == "json":
            with open(path, 'w') as f:
                json.dump(data, f, indent=4)
        elif format == "tsv":
            raise NotImplementedError("File format not supported")
        else:
            raise NotImplementedError("File format not supported")






from digitaltwins.gen3.querier import Querier


class Dataset(object):
    def __init__(self, id, program, project, config_file):
        self._program = program
        self._project = project
        self._id = id # submitter_id

        self._configs = config_file
        self._querier = Querier(self._configs)

    def get_program(self):
        return self._program

    def get_project(self):
        return self._project

    def get_id(self):
        return self._id

    def get_metadata(self, metadata):
        fields = self._get_feilds(metadata)

        if metadata == "dataset_description":
            metadata = "dataset_descriptions"
        elif metadata == "subjects":
            metadata = "cases"

        query_string = f"""
        {{
            experiment (submitter_id: "{self._id}") {{
                {metadata}{{
                    {' '.join(fields)}
                }}
            }}
        }}
        """
        metadata = self._querier.graphql_query(query_string=query_string).get("experiment")[0].get(metadata)

        return metadata

    def _get_feilds(self, metadata):
        # TODO. load from schema
        fields = list()
        if metadata == "dataset_description":
            fields = [
                "metadata_version",
                "dataset_type",
                "title",
                "subtitle",
                "keywords",
                "funding",
                "acknowledgments",
                "study_purpose",
                "study_data_collection",
                "study_primary_conclusion",
                "study_organ_system",
                "study_approach",
                "study_technique",
                "study_collection_title",
                "contributor_name",
                "contributor_orcid",
                "contributor_affiliation",
                "contributor_role",
                "identifier_description",
                "relation_type",
                "identifier",
                "identifier_type",
                "number_of_subjects",
                "number_of_samples",
                "dataset_type",
                "title",
                "subtitle",
                "keywords",
                "funding",
                "acknowledgments",
                "study_purpose",
                "study_data_collection",
                "study_primary_conclusion",
                "study_organ_system",
                "study_approach",
                "study_technique",
                "study_collection_title",
                "contributor_name",
                "contributor_orcid",
                "contributor_affiliation",
                "contributor_role",
                "identifier_description",
                "relation_type",
                "identifier",
                "identifier_type",
                "number_of_subjects",
                "number_of_samples"
            ]
        elif metadata == "subjects":
            fields = [
                "subject_id",
                "pool_id",
                "subject_experimental_group",
                "age",
                "sex",
                "species",
                "strain",
                "rrid_for_strain",
                "age_category",
                "also_in_dataset",
                "member_of",
                "laboratory_internal_id",
                "date_of_birth",
                "age_range_min",
                "age_range_max",
                "body_mass",
                "genotype",
                "phenotype",
                "handedness",
                "reference_atlas",
                "experimental_log_file_path",
                "experiment_date",
                "disease_or_disorder",
                "intervention",
                "disease_model",
                "protocol_title",
                "protocol_url_or_doi"
            ]

        return fields

        # query_string = f"""
        #         {{
        #             __type(name: "{metadata}"){{
        #                 fields{{
        #                     name
        #                 }}
        #             }}
        #         }}
        #
        #         """
        # fields_dict_list = self._querier.graphql_query(query_string=query_string).get("__type").get("fields")
        # fields = list()
        # for fields_dict in fields_dict_list:
        #     name = fields_dict.get("name")
        #     fields.append(name)




import configparser
import os
import time
from pathlib import Path

from gen3.submission import Gen3Submission
from gen3.auth import Gen3Auth, Gen3AuthError

from requests.exceptions import ConnectionError

import urllib3
urllib3.disable_warnings()


class Querier(object):
    """
    Class for querying Gen3.
    Also accepts queries in GraphQL syntax.
    """
    def __init__(self, config_file):
        """
        Constructor

        :param auth: Gen3 authentication object created by the Auth class
        :type auth: object
        """
        config_file = Path(config_file)
        self._config_file = config_file
        self._configs = configparser.ConfigParser()
        self._configs.read(config_file)

        self._config_dir = self._config_file.parent
        self._cred_file = Path(self._configs["gen3"].get("cred_file"))
        self._ssl_cert = self._configs["gen3"].get("ssl_cert")
        if self._cred_file:
            self._cred_file = self._config_dir.joinpath(self._cred_file)
        if self._ssl_cert:
            self._ssl_cert = self._config_dir.joinpath(self._ssl_cert)
            os.environ["REQUESTS_CA_BUNDLE"] = str(self._ssl_cert.resolve())


        self._endpoint = self._configs["gen3"].get("endpoint")
        self._program = self._configs["gen3"].get("program")
        self._project = self._configs["gen3"].get("project")

        self._auth = Gen3Auth(self._endpoint, str(self._cred_file))

        self._querier = Gen3Submission(self._auth)

        self._MAX_ATTEMPTS = 10

    def _get_project_id(self, program, project):
        if program is None:
            program = self._program
        if project is None:
            project = self._project

        project_id = program + '-' + project
        return project_id

    def get_program(self):
        return self._program

    def get_project(self):
        return self._project

    def graphql_query(self, query_string, variables=None, count=0):
        """
        Sending a GraphQL query to Gen3

        :param query_string: query in GraphQL syntax
        :type query_string: string
        :param variables: query variables (optional)
        :type variables: dict
        :return: query response
        :rtype: dict
        """
        if count >= self._MAX_ATTEMPTS:
            raise ValueError(f"Max attempts {count} exceeded. Please try again. If the error "
                             f"persists, please contact the developers".format(count=count))
        try:
            print("Sending request...")
            response = self._querier.query(query_string, variables)
            data = response.get("data")
            return data
        except Gen3AuthError as e:
            time.sleep(2)
            count = count + 1
            print("Connection failed.")
            return self.graphql_query(query_string, variables=variables, count=count)
        except ConnectionError as e:
            print("Connection failed.")
            raise ConnectionError("HTTP connection error: Please make sure you have access to the remote server. then "
                                  "try again!")

    def get_all_programs(self):
        """
        Getting all programs that the user have access to

        :return: List of programs
        :rtype: list
        """
        query_string = f"""
        {{
            program{{
                name
            }}
        }}
        """
        data = self.graphql_query(query_string)
        programs = data.get('program')

        return programs

    def get_projects_by_program(self, program):
        """
        Getting the projects by program name

        :param program: Name of a Gen3 program
        :type program: str
        :return: List of projects
        :rtype: list
        """
        query_string = f"""
        {{
            program (name: "{program}"){{
                name
                projects{{
                    name
                }}
            }}
        }}
        """
        data = self.graphql_query(query_string)

        projects = None
        programs = data.get('program')
        if programs and len(programs) >= 0:
            projects = programs[0].get("projects")

        return projects

    def get_datasets(self, program=None, project=None):
        if program is None and project is None:
            query_string = f"""
            {{
                program{{
                    name
                    projects{{
                        name
                        experiments{{
                            submitter_id
                        }}
                    }}
                }}
            }}
            """
        if program and project is None:
            query_string = f"""
            {{
                program (name: "{program}"){{
                    name
                    projects{{
                        name
                        experiments{{
                            submitter_id
                        }}
                    }}
                }}
            }}
            """
        if program is None and project:
            query_string = f"""
            {{
                program{{
                    name
                    projects (name: "{project}"){{
                        name
                        experiments{{
                            submitter_id
                        }}
                    }}
                }}
            }}
            """
        if program and project:
            query_string = f"""
            {{
                program (name: "{program}") {{
                    name
                    projects (name: "{project}"){{
                        name
                        experiments {{
                            submitter_id
                        }}
                    }}
                }}
            }}
            """
        data = self.graphql_query(query_string)

        datasets = list()
        programs = data.get('program')
        for program in programs:
            program_name = program.get("name")
            projects = program.get("projects")
            for project in projects:
                project_name = project.get("name")
                experiments = project.get("experiments")
                for experiment in experiments:
                    submitter_id = experiment.get("submitter_id")
                    dataset = Dataset(id=submitter_id, program=program_name, project=project_name, config_file=self._config_file)
                    datasets.append(dataset)

        return datasets

    def get_dataset(self, dataset_id, program=None, project=None):
        if program is None:
            program = self._program
        if project is None:
            project = self._project

        query_string = f"""
        {{
            program (name: "{program}"){{
                name
                projects (name: "{project}"){{
                    name
                    experiments (submitter_id: "{dataset_id}"){{
                        submitter_id
                    }}
                }}
            }}
        }}
        """
        data = self.graphql_query(query_string)

        dataset = None
        if data:
            dataset = Dataset(dataset_id, program, project, self._config_file)
        else:
            print("Dataset not found: " + str(dataset_id))

        return dataset


    def get_subjects(self, dataset_id, program=None, project=None):
        project_id = self._get_project_id(program, project)

        query_string = f"""
        {{
          experiment(project_id: "{project_id}", submitter_id: "{dataset_id}"){{
            cases{{
              subject_id
            }}
          }}
        }}
        """
        response = self.graphql_query(query_string)
        experiment = response.get("experiment")[0]
        cases = experiment.get("cases")
        return cases

    def get_dataset_descriptions(self, dataset_id, program=None, project=None):
        project_id = self._get_project_id(program, project)
        query_string = f"""
        {{
          experiment(project_id: "{project_id}", submitter_id: "{dataset_id}"){{
            dataset_descriptions{{
              metadata_version,
              dataset_type,
              title,
              subtitle,
              keywords,
              funding,
              acknowledgments,
              study_purpose,
              study_data_collection,
              study_primary_conclusion,
              study_organ_system,
              study_approach,
              study_technique,
              study_collection_title,
              contributor_name,
              contributor_orcid,
              contributor_affiliation,
              contributor_role,
              identifier_description,
              relation_type,
              identifier,
              identifier_type,
              number_of_subjects,
              number_of_samples,
              dataset_type,
              title,
              subtitle,
              keywords,
              funding,
              acknowledgments,
              study_purpose,
              study_data_collection,
              study_primary_conclusion,
              study_organ_system,
              study_approach,
              study_technique,
              study_collection_title,
              contributor_name,
              contributor_orcid,
              contributor_affiliation,
              contributor_role,
              identifier_description,
              relation_type,
              identifier,
              identifier_type,
              number_of_subjects,
              number_of_samples
            }}
          }}
        }}
        """

        response = self.graphql_query(query_string)
        dataset = response.get("experiment")[0]
        dataset_descriptions = dataset.get("dataset_descriptions")

        return dataset_descriptions

    def get_node_records(self, node, program, project):
        """
        Getting all the records in a Gen3 node

        :param node: Name of the target node
        :type node: str
        :param program: program name
        :type program: str
        :param project: project name
        :type project: str
        :return: A list of records in dictionary format
        :rtype: list
        """
        response = self._querier.export_node(program, project, node, "json")
        data = response.get("data")
        return data

    def get_dataset_records(self, dataset_id, program=None, project=None):
        project_id = self._get_project_id(program, project)

        query_string = f"""
                {{
                  experiment(project_id: "{project_id}", submitter_id: "{dataset_id}"){{
                    id,
                    cases{{
                      id,
                      samples{{
                        id,
                      }}
                    }},
                    dataset_descriptions{{
                      id,
                    }},
                    manifests{{
                      id
                    }}
                  }}
                }}
                """
        response = self.graphql_query(query_string)
        datasets = response.get("experiment")
        if len(datasets) == 0:
            return None

        records = list()
        uuid_tag = "id"
        # uuid_tag = "submitter_id"

        dataset = datasets[0]
        records.insert(0, dataset.get(uuid_tag))

        cases = dataset.get("cases")
        for case in cases:
            records.insert(0, case.get(uuid_tag))
            samples = case.get("samples")
            for sample in samples:
                records.insert(0, sample.get(uuid_tag))

        dataset_descriptions = dataset.get("dataset_descriptions")
        for dataset_desc in dataset_descriptions:
            records.insert(0, dataset_desc.get(uuid_tag))

        manifests = dataset.get("manifests")
        for manifest in manifests:
            records.insert(0, manifest.get(uuid_tag))

        return records




from digitaltwins.core.dataset import Dataset



import os
from pathlib import Path
import pandas as pd
from xlrd import XLRDError
import json
import yaml
import re
from datetime import datetime


class MetadataConvertor(object):
    """
    Converting the metadata from SPARC dataset structure (SDS) to Gen3 submittable structure in json format
    """
    def __init__(self, program, project, experiment, schema_dir=None, version="2.0.0"):
        """
        Constructor

        :param program: Program name
        :type program: str
        :param project: Project name
        :type project: str
        :param experiment: Experiment/dataset name
        :type experiment: str
        :param version: SDS schema version
        :type version: str
        """
        self._version = version
        self._program = program
        self._project = project
        self._experiment = experiment

        self._current_dir = Path(__file__).parent.resolve()

        self._resources_dir = self._current_dir.joinpath("../resources")

        if schema_dir:
            self._schema_dir = schema_dir
        else:
            version_dirname = "version_" + version.replace('.', '_')
            self._schema_dir = self._resources_dir.joinpath(version_dirname, "sds_dictionary")

        self._supported_versions = ["1.2.3", "2.0.0"]

        self._special_chars = ['/', '_']

        self._categories = ["experiment", "dataset_description", "subjects", "manifest", "samples"]
        self._row_based = ["subjects", "manifest", "samples"]
        self._col_based = ["dataset_description"]

        self._validate_version(version)

    def set_schema_dir(self, path):
        """
        Setting the SDS schema directory

        :param path: Path to the SDS schema directory
        :type path: str
        :return:
        :rtype:
        """
        self._schema_dir = Path(path)

    def _validate_version(self, version):
        """
        Checking if the SDS version is supported

        :param version: SDS version
        :type version: str
        :return:
        :rtype:
        """
        if version not in self._supported_versions:
            raise Exception("Dataset version not supported")

    def _init_data(self, category):
        """
        Initialising the Gen3 data structure

        :param category: SDS metadata category
        :type category: str
        :return: Gen3 data structure
        :rtype: dict
        """
        if category == "subjects":
            type = "case"
        else:
            type = category

        if category == "experiment":
            data = {
                "type": type,
                "submitter_id": self._experiment,
                "projects": [{
                    "code": self._project
                }],
            }
        else:
            data = {
                "type": type,
                "experiments": [{"submitter_id": self._experiment}],
                "submitter_id": self._experiment + '-' + category
            }
        return data

    def execute(self, source_dir, dest_dir):
        """
        Converting metadata

        :param source_dir: Path to the source (SDS) directory
        :type source_dir: str or  pathlib.Path object
        :param dest_dir: Path to the destination (Gen3) directory
        :type dest_dir: str or pathlib.Path object
        :return:
        :rtype:
        """
        source_dir = Path(source_dir)
        dest_dir = Path(dest_dir)
        for category in self._categories:
            data = self._init_data(category)
            if category == "experiment":
                data = self._init_data(category)
                filename = category + ".json"
                dest = dest_dir.joinpath(filename)
                self._save(data, dest)
                continue

            mappings = self._get_mappings(self._version, category)

            if category == "manifest":
                sources = self._get_files(source_dir, category)
                metadata_pd_composite = None
                for source in sources:
                    metadata_pd = self._get_sparc_metadata(source)
                    metadata_pd = self._map_fields(category, metadata_pd, mappings, target_version="2.0.0")

                    if metadata_pd_composite is not None:
                        metadata_pd_composite = pd.concat([metadata_pd_composite, metadata_pd], ignore_index=True)
                    else:
                        metadata_pd_composite = metadata_pd
                metadata_pd = metadata_pd_composite
            else:
                source = source_dir.joinpath(category + ".xlsx")
                metadata_pd = self._get_sparc_metadata(source)
                metadata_pd = self._map_fields(category, metadata_pd, mappings, target_version="2.0.0")
            records = metadata_pd.to_dict('records')

            if category in self._row_based:
                schema = self._get_schema(category)
                properties = schema.get("properties")
                system_properties = schema.get("systemProperties")
                required = schema.get("required")

                data_list = list()
                data_init = data.copy()
                for metadata_dict in records:
                    data = data_init.copy()
                    for property in properties.items():
                        key = property[0]

                        if key in system_properties:
                            continue
                        try:
                            if data.get(key):
                                continue
                        except:
                            print("test")

                        value = metadata_dict.get(key)

                        # Update submitter_id
                        if key in ["subject_id", "filename"]:
                            data["submitter_id"] = data["submitter_id"] + '-' + value
                            data["submitter_id"] = data["submitter_id"].replace('./', '')
                            data["submitter_id"] = re.sub(str(self._special_chars), '-', data["submitter_id"])

                        # check if value exists and if value equals to nan (nan variable does not equal to itself)
                        if value and value == value:
                            if isinstance(value, datetime):
                                value = str(value)

                            if isinstance(value, list) and len(value) == 1:
                                value = value[0]

                            # handle special values
                            if isinstance(value, str):
                                value = value.replace("\"", "\'")

                                # separate string by new line "\n" and saved in a list if the list > 1
                                value_list = value.split("\n")
                                if len(value_list) > 1:
                                    value = value_list

                            data[key] = value
                        else:
                            if key in required:
                                data[key] = "NA"

                    data_list.append(data)
                    del data
                data = data_list
            elif category in self._col_based:
                data = self._convert(category, records, data)

            filename = category + ".json"
            dest = dest_dir.joinpath(filename)
            self._save(data, dest)

    def _get_files(self, source, category):
        """
        Getting all files from a directory filtered by Gen3 category. e.g. manifest

        :param source: path to the source directory
        :type source: str or pathlib.Path object
        :param category: SDS category
        :type category: str
        :return: List of files
        :rtype: list
        """
        files = list()
        if source.is_file():
            files.append(source)
        elif source.is_dir():
            filename = category + ".xlsx"
            for file in source.rglob(filename):
                files.append(file)
        else:
            raise FileNotFoundError("File not found")
        return files

    @staticmethod
    def read_excel(path, sheet_name=None):
        """
        Reading Excel data as a python dataframe object

        :param path: Path to the Excel file
        :type path: str or pathlib.Path object
        :param sheet_name: Excel sheet name
        :type sheet_name: str
        :return: Data in dataframe object format
        :rtype: object
        """
        try:
            # the read_excel method return dict when sheet name is passed. otherwise a dataframe will be returned
            if sheet_name:
                metadata = pd.read_excel(path, sheet_name=sheet_name)
            else:
                metadata = pd.read_excel(path)
        except XLRDError:
            if sheet_name:
                metadata = pd.read_excel(path, sheet_name=sheet_name, engine='openpyxl')
            else:
                metadata = pd.read_excel(path, engine='openpyxl')

        return metadata

    def _get_mappings(self, version, category):
        """
        Getting the mapping between SDS and Gen3 variables

        :param version: SDS version
        :type version: str
        :param category: SDS category
        :type category: str
        :return: The mapping between SDS and Gen3 variables in dataframe
        :rtype: object
        """
        version = version.replace(".", "_")
        version = "version_" + version
        version_dir = self._resources_dir / version
        mapping_file = version_dir / "element_mapping.xlsx"

        mappings = self.read_excel(mapping_file, category)

        return mappings

    def _save(self, data, dest):
        """
        Saving the converted Gen3 submission file

        :param data: Converted data in Gen3 submission structure
        :type data: dict
        :param dest: Path to the save file
        :type dest: str or pathlib.Path object
        :return:
        :rtype:
        """
        os.makedirs(dest.parent, exist_ok=True)
        with open(dest, 'w') as f:
            json.dump(data, f, indent=4)

    def _get_schema(self, category):
        """
        Getting the SDS schema

        :param category: SDS category
        :type category: str
        :return: SDS schema
        :rtype: dict
        """
        if category == "subjects":
            category = "case"
        if category == "samples":
            category = "sample"
        schema_file = category + ".yaml"
        schema_file = self._schema_dir / schema_file

        try:
            # Linux
            with open(schema_file, 'r') as yml:
                schema = yaml.safe_load(yml)
        except Exception as e:
            # Windows
            with open(schema_file, 'rt', encoding='utf8') as yml:
                schema = yaml.safe_load(yml)

        return schema

    def _get_sparc_metadata(self, source):
        """
        Getting the SDS metadata from a file (Excel)

        :param source: Path to the metadata file
        :type source: str
        :return:
        :rtype:
        """
        metadata_sparc = self.read_excel(source)
        if source.stem == "dataset_description":
            # combine values
            metadata_sparc["values"] = metadata_sparc.iloc[:, 3:].values.tolist()
            metadata_sparc["values"] = metadata_sparc["values"].apply(lambda x: [i for i in x if str(i) != "nan"])

            # Extract columns
            # Element name: 1st column
            # Values: last column
            metadata_sparc = metadata_sparc[[metadata_sparc.columns[0], metadata_sparc.columns[-1]]]

        return metadata_sparc

    def _map_fields(self, category, metadata_sparc, mappings, target_version):
        """
        Mapping/changing the SDS fields' names to Gen3 fields' names

        :param category: SDS category
        :type category: str
        :param metadata_sparc: SDS metadata in Dataframe
        :type metadata_sparc: object
        :param mappings: Mapping in Dataframe
        :type mappings: object
        :param target_version: SDS version
        :type target_version: str
        :return: New SDS metadata in Dataframe
        :rtype: object
        """
        if category == "dataset_description":
            nums_of_records = len(metadata_sparc)
            column_idx = 0
            elements = metadata_sparc.iloc[:, column_idx].tolist()

            for idx in range(nums_of_records):
                element = elements[idx]
                try:
                    record = mappings[mappings['element'] == element]
                    metadata_sparc.loc[idx, metadata_sparc.columns[0]] = record.iloc[0]["gen3_element_" + target_version]
                except:
                    continue

            metadata_sparc = metadata_sparc.dropna()
        else:
            column_headers = list(metadata_sparc.columns)
            for idx, column_header in enumerate(column_headers):
                record = mappings[mappings['element'].str.lower() == column_header.lower()]
                if record.empty:
                    continue
                new_name = record.iloc[0]["gen3_element_" + target_version]

                metadata_sparc = metadata_sparc.rename(columns={column_header: new_name})

            metadata_sparc = metadata_sparc.loc[:, metadata_sparc.columns.notna()]

        return metadata_sparc

    def _convert(self, category, records, data):
        """
        Converting SDS data to Gen3 submission structure

        :param category: SDS category
        :type category: str
        :param records: SDS data. List of dictionary
        :type records: list
        :param data: Initial data in Gen3 structure
        :type data: dict
        :return: New data in Gen3 structure
        :rtype: dict
        """
        metadata_dict = dict()

        if category == "dataset_description":
            for record in records:
                element = record.get("Metadata element")
                values = record.get("values")
                metadata_dict[element] = values

        schema = self._get_schema(category)
        properties = schema.get("properties")
        system_properties = schema.get("systemProperties")
        required = schema.get("required")

        for property in properties.items():
            key = property[0]

            if key in system_properties:
                continue
            if data.get(key):
                continue

            value = metadata_dict.get(key)

            if value:
                if category in self._row_based and len(value) == 1:
                    value = value[0]

                data[key] = value
            else:
                if key in required:
                    if category in self._col_based:
                        data[key] = []
                    else:
                        data[key] = "NA"
        return data



import configparser
import os
import re
import time

from pathlib import Path
import shutil

from digitaltwins import Querier
from digitaltwins import MetadataConvertor
from digitaltwins import MetadataUploader
from digitaltwins.irods.irods import IRODS

import urllib3
urllib3.disable_warnings()


class Uploader(object):
    def __init__(self, config_file):
        config_file = Path(config_file)
        self._config_file = config_file
        self._configs = configparser.ConfigParser()
        self._configs.read(config_file)

        self._config_dir = self._config_file.parent
        self._gen3_cred_file = Path(self._configs["gen3"].get("cred_file"))
        self._ssl_cert = self._configs["gen3"].get("ssl_cert")
        if self._gen3_cred_file:
            self._gen3_cred_file = self._config_dir.joinpath(self._gen3_cred_file)
        if self._ssl_cert:
            self._ssl_cert = self._config_dir.joinpath(self._ssl_cert)
            os.environ["REQUESTS_CA_BUNDLE"] = str(self._ssl_cert.resolve())

        self._gen3_endpoint = self._configs["gen3"].get("endpoint")

        self._program = self._configs["gen3"].get("program")
        self._project = self._configs["gen3"].get("project")

        time_stamp = time.time()
        dirname_tmp = "./tmp_{}".format(time_stamp)
        self._dir_tmp = Path(dirname_tmp)
        self._meta_dir_tmp = self._dir_tmp.joinpath(Path("gen3_tmp"))

        self._meta_files = ["experiment.json", "dataset_description.json", "manifest.json", "subjects.json"]

        self._dataset_submitter_id_template = "{program}-{project}-dataset-{id}-version-1"
        self._dataset_id_index = 3
        self._MAX_ATTEMPTS = 10

    def upload(self, dataset_dir):
        dataset_dir = Path(dataset_dir)
        self._verify_dataset(dataset_dir)

        # get dataset (submitter) id
        dataset_id = self._generate_dataset_id()

        os.makedirs(self._dir_tmp, exist_ok=True)

        dataset_dir_tmp = self._dir_tmp.joinpath(dataset_id)
        shutil.copytree(str(dataset_dir), str(dataset_dir_tmp))
        dataset_dir = dataset_dir_tmp

        # Upload metadata to Gen3
        self.upload_metadata(dataset_dir)
        # Upload the actual files to iRODS
        self.upload_dataset(dataset_dir)

        if self._dir_tmp.is_dir:
            shutil.rmtree(str(self._dir_tmp))

        print("Dataset uploaded: " + str(dataset_id))

    def upload_metadata(self, dataset_dir):
        meta_dir = self._meta_dir_tmp
        # convert sds metadata to gen3
        meta_convertor = MetadataConvertor(program=self._program, project=self._project, experiment=dataset_dir.name)
        meta_convertor.execute(source_dir=dataset_dir, dest_dir=meta_dir)

        # upload metadata
        meta_uploader = MetadataUploader(self._gen3_endpoint, str(self._gen3_cred_file))

        for filename in self._meta_files:
            print("Uploading: " + str(filename))
            file = meta_dir.joinpath(filename)
            meta_uploader.execute(program=self._program, project=self._project, file=str(file))

        # delete the temporary metadata dir
        if meta_dir.is_dir:
            shutil.rmtree(meta_dir)

    def upload_dataset(self, dataset_dir):
        irods = IRODS(self._configs)
        irods.upload(dataset_dir)

    def _verify_dataset(self, dataset_dir):
        if dataset_dir.is_dir():
            dataset_name = dataset_dir.name
        else:
            raise NotADirectoryError("Dataset directory not found")

        # check if dataset exists
        # todo
        pass

    def _generate_dataset_id(self, count=0):
        if count >= self._MAX_ATTEMPTS:
            raise ValueError("Max attempts {count} exceeded. Please try submitting again. If the error persists, "
                             "please contact the developers".format(count=count))
        # list datasets
        querier = Querier(self._config_file)

        datasets = list()
        try:
            datasets = querier.get_datasets(program=self._program, project=self._project)
        except Exception:
            time.sleep(2)
            self._generate_dataset_id(count=count + 1)

        dataset_ids = list()
        for dataset in datasets:
            id = dataset.get_id()
            dataset_ids.append(id)

        if len(datasets) > 0:
            dataset_ids.sort()
            latest_dataset = dataset_ids[-1]
            elements = re.split('_|-', latest_dataset)
            latest_id = elements[self._dataset_id_index]
            new_id = int(latest_id) + 1
            new_dataset_id = self._dataset_submitter_id_template.format(program=self._program, project=self._project, id=new_id)
        else:
            return self._dataset_submitter_id_template.format(program=self._program, project=self._project, id="1")

        return new_dataset_id



import json
import time

from gen3.auth import Gen3Auth, Gen3AuthError
from gen3.submission import Gen3Submission

import urllib3
urllib3.disable_warnings()


class MetadataUploader(object):
    """
    Class for Gen3 submission
    """

    def __init__(self, endpoint, credentials):
        """
        Constructor

        :param endpoint: Gen3 root URL
        :type endpoint: str
        :param credentials: Path to a Gen3 credentials file
        :type credentials: str
        """
        self._endpoint = endpoint
        self._credentials = credentials

        self._auth = Gen3Auth(endpoint, refresh_file=self._credentials)
        self._submission = Gen3Submission(endpoint, self._auth)

        self._MAX_ATTEMPTS = 10

    def submit(self, program, project, record, count):
        if count >= self._MAX_ATTEMPTS:
            raise ValueError(f"Max submission attempts {count} exceeded. Please try submitting again. If the error "
                             f"persists, please contact the developers".format(count=count))
        try:
            self._submission.submit_record(program, project, record)
        except Gen3AuthError as e:
            time.sleep(2)
            self.submit(program, project, record, count + 1)

    def execute(self, program, project, file):
        """
        Submitting metadata to Gen3

        :param program: Program name
        :type program: str
        :param project: Project name
        :type project: str
        :param file: Path to the metadata file
        :type file: str
        :return:
        :rtype:
        """
        with open(file, 'r') as f:
            record = json.load(f)

        # self._submission.submit_record(program, project, record)

        self.submit(program, project, record, count=0)





from .gen3.auth import Auth
from .gen3.metadata_convertor import MetadataConvertor
from .gen3.metadata_exporter import MetadataExporter
from .gen3.querier import Querier
from .gen3.metadata_uploader import MetadataUploader
from .irods.irods import IRODS
from .core.downloader import Downloader
from .core.uploader import Uploader
from .core.deleter import Deleter
from .core.dataset import Dataset



import configparser
import os
from pathlib import Path

from digitaltwins import IRODS


class Downloader(object):
    def __init__(self, config_file):
        self._configs = configparser.ConfigParser()
        self._configs.read(config_file)

        self._save_dir = Path(r"./tmp")

        self._irods = IRODS(self._configs)

    def download(self, dataset_name=None, save_dir=None):
        if dataset_name is None:
            raise ValueError("Dataset not specified")
        if save_dir is None:
            save_dir = self._save_dir

        os.makedirs(str(save_dir), exist_ok=True)

        print("Downloading dataset " + dataset_name)

        self._irods.download(dataset_name, save_dir)

        print("Dataset successfully downloaded")

    def _download_collection(self, session, collection_path, save_dir):
        dataset = session.collections.get(collection_path)
        save_dir = save_dir.joinpath(dataset.name)
        os.makedirs(save_dir, exist_ok=True)

        for obj in dataset.data_objects:
            session.data_objects.get(obj.path, os.path.join(save_dir, obj.name))

        if dataset.subcollections:
            for subcollection in dataset.subcollections:
                self._download_collection(session, subcollection.path, save_dir)
        else:
            return



from gen3.auth import Gen3Auth


class Auth(Gen3Auth):
    """
    Class for Gen3 Authentication inherited from the offical Gen3 API's Gen3 Auth Helper
    https://gen3sdk-python.readthedocs.io/en/latest/auth.html
    """

    __doc__ += Gen3Auth.__doc__
    pass



from setuptools import setup, find_packages

# from pathlib import Path
# this_directory = Path(__file__).parent
# long_description = (this_directory / "README.md").read_text() ## UnicodeDecodeError when decoding the REAME file

setup(
    name="digitaltwins",
    version="0.3.8",
    description='A Python tool for interacting with the 12 Labours DigitalTWINS (Digital Translational Workflows for '
                'Integrating Systems) Platform',
    # long_description=long_description,
    # long_description_content_type='text/markdown',
    author="Thiranja Prasad Babarenda Gamage, Chinchien Lin, Linkun Gao, Jiali Xu, David Nickerson",
    email="psam012@aucklanduni.ac.nz, clin864@aucklanduni.ac.nz",
    license="Apache-2.0",
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['./resources/**']},
    install_requires=[
        "gen3~=4.21.0",
        "python-irodsclient>=1.1.8",
        "pandas",
        "xlrd",
        "PyYAML",
        "openpyxl",
        "jsonschema",
        "dictionaryutils"
    ]
)



import os
from pathlib import Path

from irods.session import iRODSSession

class IRODS(object):
    """
    Class for interacting with iRODS server
    """
    def __init__(self, configs):
        """
        Constructor
        """

        self._configs = configs["irods"]
        self._host = self._configs.get("irods_host")
        self._port = self._configs.get("irods_port")
        self._user = self._configs.get("irods_user")
        self._password = self._configs.get("irods_password")
        self._zone = self._configs.get("irods_zone")
        self._project_root = self._configs.get("irods_project_root")

    def upload(self, local_dataset_dir):
        with iRODSSession(host=self._host,
                          port=self._port,
                          user=self._user,
                          password=self._password,
                          zone=self._zone) as session:

            local_dataset_dir = Path(local_dataset_dir)

            self._create_collections(session, local_dataset_dir)

            try:
                # Walk through the local directory and upload its contents to iRODS
                for root, dirs, files in os.walk(str(local_dataset_dir)):
                    for file in files:
                        local_file_path = os.path.join(root, file)
                        filename = Path(local_file_path).name

                        sub = str(root).replace(str(local_dataset_dir.parent), '')
                        sub = sub.replace("\\", "/")
                        collection_path = self._project_root + sub
                        irods_file_path = collection_path + "/" + filename

                        # Upload the file to iRODS
                        session.data_objects.put(local_file_path, irods_file_path, force=True)

                        print(f"File uploaded: '{local_file_path}'")
            except Exception as e:
                print(f"Error uploading local directory: {e}")

    def _create_collections(self, session, local_dataset_dir):
        for item in local_dataset_dir.rglob("*"):
            if item.is_dir():
                sub = str(item).replace(str(local_dataset_dir.parent), '')
                sub = sub.replace("\\", "/")
                collection_path = self._project_root + sub

                if session.collections.exists(collection_path):
                    continue
                else:
                    session.collections.create(collection_path, recurse=True)


    def download(self, collection_name, save_dir=None):
        """
        Downloading data from iRODS

        :param collection_name: iRODS collection/Dataset name in the iRODS zone
        :type collection_name: str
        :param save_dir: path to the save directory
        :type save_dir: str
        :return:
        :rtype:
        """
        with iRODSSession(host=self._host,
                          port=self._port,
                          user=self._user,
                          password=self._password,
                          zone=self._zone) as session:
            collection_path = self._project_root + '/' + collection_name
            self._download_collection(session, collection_path, save_dir=save_dir)

    def _download_collection(self, session, collection_path, save_dir):
        """
        Downloading an iRODS collection

        :param session: iRODS connection session
        :type session: object
        :param collection_path: collection path in the iRODS system
        :type collection_path: str
        :param save_dir: local path to the save directory
        :type save_dir: str
        :return:
        :rtype:
        """
        collection = session.collections.get(collection_path)
        save_dir = Path(save_dir)
        save_dir = save_dir.joinpath(collection.name)
        os.makedirs(save_dir, exist_ok=True)

        for dobj in (collection.data_objects):
            session.data_objects.get(dobj.path, os.path.join(save_dir,dobj.name))

        if collection.subcollections:
           for subcollection in collection.subcollections:
               self._download_collection(session, subcollection.path, save_dir)
        else:
            return

    def delete(self, collection_name):
        with iRODSSession(host=self._host,
                          port=self._port,
                          user=self._user,
                          password=self._password,
                          zone=self._zone) as session:
            collection_path = self._project_root + '/' + collection_name

            if session.collections.exists(collection_path):
                session.collections.remove(collection_path, recurse=True)



import json


class ConfigLoader(object):
    def __init__(self, file=None):
        self._configs = None
        if file:
            self._configs = self.load_from_json(file)

    @staticmethod
    def load_from_json(file):
        with open(file) as f:
            configs = json.load(f)
        return configs



import configparser
import os
from pathlib import Path

from gen3.auth import Gen3Auth
from gen3.submission import Gen3Submission

from digitaltwins import Querier
from digitaltwins.irods.irods import IRODS

import urllib3
urllib3.disable_warnings()


class Deleter(object):
    def __init__(self, config_file):
        config_file = Path(config_file)
        self._config_file = config_file
        self._configs = configparser.ConfigParser()
        self._configs.read(config_file)


        self._config_dir = self._config_file.parent
        self._gen3_cred_file = Path(self._configs["gen3"].get("cred_file"))
        self._ssl_cert = self._configs["gen3"].get("ssl_cert")
        if self._gen3_cred_file:
            self._gen3_cred_file = self._config_dir.joinpath(self._gen3_cred_file)
        if self._ssl_cert:
            self._ssl_cert = self._config_dir.joinpath(self._ssl_cert)
            os.environ["REQUESTS_CA_BUNDLE"] = str(self._ssl_cert.resolve())

        self._program = self._configs["gen3"].get("program")
        self._project = self._configs["gen3"].get("project")
        self._gen3_endpoint = self._configs["gen3"].get("endpoint")

        self._auth = Gen3Auth(self._gen3_endpoint, refresh_file=str(self._gen3_cred_file))
        self._submission = Gen3Submission(self._gen3_endpoint, self._auth)

    def delete(self, dataset_id):
        self.delete_metadata(dataset_id)
        self.delete_dataset(dataset_id)

    def delete_metadata(self, dataset_id):
        querier = Querier(self._config_file)
        records = querier.get_dataset_records(dataset_id=dataset_id, program=self._program, project=self._project)
        self._submission.delete_records(program=self._program, project=self._project, uuids=records)

    def delete_dataset(self, dataset_id):
        irods = IRODS(self._configs)
        irods.delete(dataset_id)
